from typing import Dict, Union, List, Literal
from xml.etree import ElementTree

# https://github.com/python/typing/issues/182
JSONType = Union[str, int, float, bool, None, Dict[str, 'JSON'], List['JSON']]
DataType = Union[JSONType, ElementTree.Element]


ReturnType = Literal['json', 'xml']

EntrezDatabaseType = Literal[
    'bioproject',
    'biosample',
    'biosystems',
    'books',
    'cdd',
    'gap',
    'dbvar',
    'epigenomics',
    'nucest',
    'gene',
    'genome',
    'gds',
    'geoprofiles',
    'nucgss',
    'homologene',
    'mesh',
    'toolkit',
    'ncbisearch',
    'nlmcatalog',
    'nuccore',
    'omia',
    'popset',
    'probe',
    'protein',
    'proteinclusters',
    'pcassay',
    'pccompound',
    'pcsubstance',
    'pubmed',
    'pmc',
    'snp',
    'sra',
    'structure',
    'taxonomy',
    'unigene',
    'unists'
]


CommandType = Literal[
    'neighbor',
    'neighbor_score',
    'neighbor_history',
    'acheck',
    'ncheck',
    'lcheck',
    'llinks',
    'llinkslib',
    'prlinks'
]



from collections import defaultdict
from csv import DictReader
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List
from warnings import warn

from .types import ReturnType, EntrezDatabaseType, CommandType


def _read_table(path: Path) -> Dict[str, List]:
    """Return dict where each entry is a data column"""
    table = defaultdict(list)
    with open(path) as f:
        reader = DictReader(f, delimiter='\t')
        for row in reader:
            for field, value in row.items():
                table[field].append(value)
    return table


# https://www.ncbi.nlm.nih.gov/books/NBK25497/table/chapter2.T._entrez_unique_identifiers_ui/?report=objectonly
data_path = (Path(__file__).parent / 'data').resolve()
entrez_databases = _read_table(data_path / 'entrez_databases.tsv')

entrez_database_codes = entrez_databases['E-utility Database Name']


@dataclass
class EntrezQuery:

    database: EntrezDatabaseType
    method = 'get'
    endpoint_suffix = '.fcgi'

    def validate(self):
        if self.database not in entrez_database_codes:
            warn(f'Unknown database: {self.database}')

    @property
    def uid_meaning(self):
        return entrez_databases

    def to_params(self) -> Dict[str, str]:
        """Convert to params which can be accepted by Entrez"""
        # TODO maybe use pydantic isntead?
        self.validate()
        params = {}
        if self.database:
            params['db'] = self.database
        return params

    @property
    def summary(self):
        return f'{self.__class__.__name__} in {self.database}'


@dataclass
class InfoQuery(EntrezQuery):
    endpoint = 'einfo'


@dataclass
class SearchQuery(EntrezQuery):
    """
    Functions:
        - Provides a list of UIDs matching a text query
        - Posts the results of a search on the History server
        - Downloads all UIDs from a dataset stored on the History server
        - Combines or limits UID datasets stored on the History server
        - Sorts sets of UIDs

    Required Parameters:
        database: Database to search. Value must be a valid E-utility database name - see `entrez_database_codes` (default = `'pubmed'`).
        term: Entrez text query

    """
    endpoint = 'esearch'
    term: str
    max_results: int

    def validate(self):
        super().validate()
        if self.max_results > 100_000:
            raise ValueError('Fetching more than 100,000 results is not implemented')

    def to_params(self) -> Dict[str, str]:
        params = super().to_params()
        params['retmax'] = str(self.max_results)
        params['term'] = self.term
        return params

    @property
    def summary(self):
        return f'{self.__class__.__name__} {self.term!r} in {self.database}'


@dataclass
class SummaryQuery(EntrezQuery):
    """
    Functions:
        - Returns document summaries (DocSums) for a list of input UIDs
        - Returns DocSums for a set of UIDs stored on the Entrez History server

    Required Parameters:
        database: Database to search. Value must be a valid E-utility database name - see `entrez_database_codes` (default = `'pubmed'`).
        ids: UID list. Either a single UID or a comma-delimited list of UIDs may be provided. All of the UIDs must be from the database specified by `database`
    """
    endpoint = 'esummary'
    method = 'post'
    ids: List[str]
    max_results: int

    def validate(self):
        super().validate()
        if self.max_results > 10_000:
            raise ValueError('Fetching more than 10,000 results is not implemented')

    def to_params(self) -> Dict[str, str]:
        params = super().to_params()
        params['retmax'] = str(self.max_results)
        params['id'] = ','.join([
            id.strip()
            for id in self.ids
        ])
        return params

    @property
    def summary(self):
        ids_summary = self.ids if len(self.ids) <= 5 else f'{len(self.ids)} ids'
        return f'{self.__class__.__name__} {ids_summary} in {self.database}'


@dataclass
class FetchQuery(SummaryQuery):
    """
    It enforces xml as a default return_type as JSON is not properly implemented by the eutilis server yet.
    """
    # 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/efetch.fcgi?db=pubmed&id=11748933,11700088&retmode=xml'
    endpoint = 'efetch'
    return_type: ReturnType = 'xml'

    def to_params(self) -> Dict[str, str]:
        params = super().to_params()
        params['retmode'] = self.return_type
        return params


@dataclass
class LinkQuery(EntrezQuery):
    """
    Functions:
        Returns UIDs linked to an input set of UIDs in either the same or a different Entrez database
        Returns UIDs linked to other UIDs in the same Entrez database that match an Entrez query
        Checks for the existence of Entrez links for a set of UIDs within the same database
        Lists the available links for a UID
        Lists LinkOut URLs and attributes for a set of UIDs
        Lists hyperlinks to primary LinkOut providers for a set of UIDs
        Creates hyperlinks to the primary LinkOut provider for a single UID

    """
    endpoint = 'elink'

    database_from: EntrezDatabaseType
    command: CommandType = 'neighbor'



from functools import wraps
from math import ceil
from time import sleep
from typing import Iterable
from warnings import warn

from requests import RequestException


try:
    from tqdm import tqdm
except ImportError:
    def tqdm(iterable):
        return iterable


def batches(data, size=100):
    return [
        data[i * size:(i + 1) * size]
        for i in range(0, ceil(len(data) / size))
    ]


def suport_batches(func):
    """
    Call the decorated functions with the collection from the first argument
    (second if counting with self) split into batches, resuming on failures
    with a interval twice the between-batch interval.
    """

    @wraps(func)
    def batches_support_wrapper(self: 'EntrezAPI', collection: Iterable, *args, **kwargs):
        size = self._batch_size
        interval = self._batch_sleep_interval
        if size is not None:
            assert isinstance(size, int)
            by_batch = {}

            for i, batch in enumerate(tqdm(batches(collection, size=size))):
                done = False

                while not done:
                    reason = None
                    try:
                        batch_result = func(self, batch, *args, **kwargs)
                        code = batch_result.response.status_code
                        if code == 200:
                            done = True
                        else:
                            reason = f'Status code != 200 (= {code})'
                    except RequestException as e:
                        reason = e

                    if not done:
                        warn(
                            f'Failed to fetch for {i}-th batch, retrying in {interval * 2} seconds.'
                            f' The reason was: {reason}'
                        )
                        sleep(interval * 2)

                by_batch[tuple(batch)] = batch_result
                sleep(interval)
            return by_batch
        else:
            return func(self, collection, *args, **kwargs)

    return batches_support_wrapper



from setuptools import setup
from setuptools import find_packages


try:
    from pypandoc import convert

    def get_long_description(file_name):
        return convert(file_name, 'rst', 'md')

except ImportError:

    def get_long_description(file_name):
        with open(file_name) as f:
            return f.read()


if __name__ == '__main__':
    setup(
        name='easy_entrez',
        packages=find_packages(),
        package_data={'easy_entrez': ['data/*.tsv']},
        version='0.1.1',
        license='MIT',
        description='Python REST API for Entrez E-Utilities: stateless, easy to use, reliable.',
        long_description=get_long_description('README.md'),
        author='Michal Krassowski',
        author_email='krassowski.michal+pypi@gmail.com',
        url='https://github.com/krassowski/easy-entrez',
        keywords=['entrez', 'pubmed', 'e-utilities', 'ncbi', 'rest', 'api'],
        classifiers=[
            'Development Status :: 4 - Beta',
            'License :: OSI Approved :: MIT License',
            'Operating System :: Microsoft :: Windows',
            'Operating System :: POSIX :: Linux',
            'Topic :: Utilities',
            'Topic :: Scientific/Engineering :: Bio-Informatics',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Intended Audience :: Developers',
            'Intended Audience :: Science/Research',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7'
        ],
        install_requires=['requests'],
    )



import requests
from requests import Response
from typing import Dict, List
from xml.etree import ElementTree
from copy import copy
from time import time, sleep

from .batch import suport_batches
from .types import ReturnType, DataType, EntrezDatabaseType, CommandType
from .queries import EntrezQuery, SearchQuery, SummaryQuery, FetchQuery, LinkQuery


class EntrezResponse:

    def __init__(self, query: EntrezQuery, response: Response, api: 'EntrezAPI'):
        self.query: EntrezQuery = query
        self.response: Response = response
        self.api: 'EntrezAPI' = api

    @property
    def content_type(self) -> ReturnType:
        declared_type = self.response.headers['Content-Type']
        if declared_type.startswith('application/json'):
            return 'json'
        if declared_type.startswith('text/xml'):
            return 'xml'
        raise ValueError(f'Unknown content type: {declared_type}')

    @property
    def data(self) -> DataType:
        if self.content_type == 'json':
            return self.response.json()
        if self.content_type == 'xml':
            return ElementTree.fromstring(self.response.content)
        raise ValueError(f'Unknown data data {self.content_type}')

    def __repr__(self):
        query = self.query
        response = self.response
        return f'<EntrezResponse status={response.status_code} for {query.summary}>'


class EntrezAPI:
    server = 'https://eutils.ncbi.nlm.nih.gov/entrez/eutils/'

    def __init__(
        self, tool: str, email: str, api_key=None,
        return_type: ReturnType = 'json',
        minimal_interval: int = 1 / 3,
        timeout: int = 10
    ):
        """
        Args:
            minimal_interval: the time interval (seconds) to be enforced between consecutive requests;
              by default 1/3 of a second to comply with the Entrez guidelines,
              but you may increase it if you want to be kind to others,
              or decrease it if you have an API key with an appropriate consent from Entrez.
        """
        self.tool = tool
        self.email = email
        self.api_key = api_key
        self.return_type = return_type
        self.minimal_interval = minimal_interval
        self._batch_size: int = None
        self._batch_sleep_interval: int = 3
        self._last_request_time = None
        self.timeout = timeout

    def _base_params(self) -> Dict[str, str]:
        return {
            'tool': self.tool,
            'email': self.email,
            'api_key': self.api_key,
            'retmode': self.return_type
        }

    def _request(self, query: EntrezQuery, custom_payload=None) -> EntrezResponse:
        url = f'{self.server}{query.endpoint}{query.endpoint_suffix}'

        data = {
            # TODO maybe warn if overwriting?
            **self._base_params(),
            **query.to_params(),
            **(custom_payload or {})
        }

        current_time = time()
        if self._last_request_time is not None:
            elapsed = current_time - self._last_request_time
            if elapsed < self.minimal_interval:
                to_wait = self.minimal_interval - elapsed
                sleep(to_wait)
        self._last_request_time = current_time

        if query.method == 'get':
            response = requests.get(url, params=data, timeout=self.timeout)
        elif query.method == 'post':
            response = requests.post(url, data=data, timeout=self.timeout)
        else:
            raise ValueError(f'Incorrect query method: {query.method}')

        return EntrezResponse(query=query, response=response, api=self)

    def search(
        self, term: str, max_results: int,
        database: EntrezDatabaseType = 'pubmed', min_date=None, max_date=None
    ):
        assert not min_date and not max_date  # TODO
        query = SearchQuery(term=term, max_results=max_results, database=database)
        return self._request(query=query)

    def in_batches_of(self, size: int = 100, sleep_interval: int = 3):
        batch_mode = copy(self)
        batch_mode._batch_size = size
        batch_mode._batch_sleep_interval = sleep_interval
        return batch_mode

    @suport_batches
    def summarize(
        self, ids: List[str], max_results: int,
        database: EntrezDatabaseType = 'pubmed'
    ):
        query = SummaryQuery(ids=ids, max_results=max_results, database=database)
        return self._request(query=query)

    @suport_batches
    def fetch(
        self, ids: List[str], max_results: int,
        database: EntrezDatabaseType = 'pubmed', return_type: ReturnType = 'xml'
    ):
        query = FetchQuery(ids=ids, max_results=max_results, database=database, return_type=return_type)
        return self._request(query=query)

    @suport_batches
    def link(
        self,
        # required
        ids: List[str],
        database_to: EntrezDatabaseType,
        database_from: EntrezDatabaseType,
        # optional
        command: CommandType = 'neighbor'
    ):
        query = LinkQuery(
            ids=ids, database=database_to, database_from=database_from,
            command=command
        )
        return self._request(query=query)
